#===============================================================================
#   SpiderLing -- A web spider/crawler for linguistic purpose.
#   Copyright (C) 2011-2015  Vit Suchomel
#   http://corpus.tools/wiki/SpiderLing
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================

import sys
import select
from time import sleep
from threading import Thread

import util
from util import http
from util import config

import logging
logging.basicConfig(filename='%s.log.crawl' % config.PREFIX, filemode='a',
    level=config.LOG_LEVEL, format='%(asctime)-15s CRAWL %(message)s')

def send_redirect_urls(urls):
    url_pairs = [u'%s\t%s' % (src_host, url) for src_host, url in urls]
    data = u'%s\n%s\n' % (util.MSG_NEW_URL, u'\n'.join(url_pairs))
    util.send_data(sys.stdout, data)

def send_robot_lines(host, robot_lines):
    if not host:
        return
    if util.MSG_ROBOT_FAIL != robot_lines:
        try:
            robot_lines = unicode(robot_lines, 'utf-8')
        except UnicodeDecodeError:
            try:
                robot_lines = unicode(robot_lines, 'iso-8859-1')
            except UnicodeDecodeError:
                logging.debug('Could not decode robots for host %s' % host.encode('utf-8'))
                robot_lines = util.MSG_ROBOT_FAIL
    robot_msg = u'%s\n%s\n%s' % (util.MSG_ROBOT_LINES, host, robot_lines)
    util.send_data(sys.stdout, robot_msg)

def receive_data(url_queue, q_docs, terminate):
    FULL_QUEUE_DELAY = 1.0
    #wait for stdin to read urls
    while True:
        #do not read more when the input or the output queue is full
        url_queue.acquire()
        url_queue_len = len(url_queue)
        url_queue.release()
        if config.MAX_URL_QUEUE <= url_queue_len or config.MAX_DOC_QUEUE < len(q_docs):
            sleep(FULL_QUEUE_DELAY)
            continue
        #read data
        input_data = util.recv_data(sys.stdin)
        #detect termination order
        if util.TERM_ORDER == input_data:
            logging.info('TERMINATE')
            terminate[0] = True
            #send out a notice we acknowledge the request to stop
            util.send_data(sys.stdout, util.TERM_RESPONSE, flush=True)
            break

        #update the url queue
        url_queue.acquire()
        #detect dead domains message
        if util.MSG_DEAD_DOMS in input_data:
            doms_dead = input_data.strip().split('\t')[1:]
            url_queue_new = []
            for u in url_queue:
                if not u[2] in doms_dead:
                    url_queue_new.append(u)
            url_queue.clear()
            url_queue.extend(url_queue_new)
        #read new urls to download
        else:
            for row in input_data.splitlines():
                try:
                    schema, host, path, ip, robot_flag = row.split('\t')
                except ValueError:
                    logging.warning('Ignoring incorrect url "%s"' % row.encode('utf-8'))
                    continue
                url = http.url_join(schema, host, path)
                is_robot = (util.ROBOT_FLAG == robot_flag)
                url_queue.append((url, schema, host, path, ip, is_robot))
        url_queue.release()

def create_connections(url_queue, epoll, q_robots, conns_write, conns_read,
        q_docs, terminate):
    MAX_CONN_DELAY = 0.1
    MAX_DOC_QUEUE_DELAY = 1.0
    NOT_ALL_OPENED_DELAY = 0.1
    LAST_CONN_TIME_CLEANUP_PERIOD = 300
    MAX_LAST_CONN_TIME = 60
    host_last_connection_time = {}
    ip_last_connection_time = {}
    last_conn_time_cleanup_time = util.now()
    while not terminate[0]:
        #full document queue => wait
        if config.MAX_DOC_QUEUE < len(q_docs):
            sleep(MAX_DOC_QUEUE_DELAY)
            continue
        #max connections reached => wait
        if config.MAX_OPEN_CONNS <= len(conns_write) + len(conns_read) + config.OPEN_AT_ONCE:
            sleep(MAX_CONN_DELAY)
            continue
        #open new connections
        conns_skipped_host = 0
        conns_skipped_ip = 0
        opened = 0
        requeue = set()
        url_queue.acquire()
        while config.OPEN_AT_ONCE > opened:
            #pop a new url from the queue
            try:
                url_parts = url_queue.popleft()
            except IndexError:
                #empty url queue
                break
            #host connection interval check
            host = url_parts[2]
            check_time = util.now()
            try:
                check_fail = (config.HOST_CONN_INTERVAL > util.seconds_total(
                    check_time - host_last_connection_time[host]))
            except KeyError:
                check_fail = False
            if check_fail:
                #prevent connecting too early
                requeue.add(url_parts)
                conns_skipped_host += 1
                continue
            #IP connection interval check
            ip = url_parts[4]
            try:
                check_fail = (config.IP_CONN_INTERVAL > util.seconds_total(
                    check_time - ip_last_connection_time[ip]))
            except KeyError:
                check_fail = False
            if check_fail:
                #prevent connecting too early
                requeue.add(url_parts)
                conns_skipped_ip += 1
                continue
            host_last_connection_time[host] = check_time
            ip_last_connection_time[ip] = check_time

            #create & open a new connection
            # non blocking -- may not connect immediately
            opened += 1
            conn = http.Connection(*url_parts)
            try:
                conn.connect()
            except http.ConnectionException:
                if conn.is_robot():
                    q_robots.add_((conn.get_host(), util.MSG_ROBOT_FAIL))
                continue
            #register the connection and add it to the write queue
            conn_id = conn.fileno()
            epoll.register(conn_id, select.EPOLLOUT)
            conns_write.set_(conn_id, conn)

        url_queue.extend(requeue)
        url_queue.release()
        logging.debug('conns opened: %d, waiting (host/IP): %d/%d' % 
            (opened, conns_skipped_host, conns_skipped_ip))

        #remove last connection time records for idle ip/hosts
        if util.seconds_left(last_conn_time_cleanup_time, LAST_CONN_TIME_CLEANUP_PERIOD) < 1:
            host_last_connection_time_new = {}
            for host, last_conn_time in host_last_connection_time.iteritems():
                if util.seconds_left(last_conn_time, MAX_LAST_CONN_TIME) > 0:
                    host_last_connection_time_new[host] = last_conn_time
            host_last_connection_time = host_last_connection_time_new
            ip_last_connection_time_new = {}
            for ip, last_conn_time in host_last_connection_time.iteritems():
                if util.seconds_left(last_conn_time, MAX_LAST_CONN_TIME) > 0:
                    ip_last_connection_time_new[ip] = last_conn_time
            ip_last_connection_time = ip_last_connection_time_new
            last_conn_time_cleanup_time = util.now()
        #wait if not at full load:
        if config.OPEN_AT_ONCE != opened:
            sleep(NOT_ALL_OPENED_DELAY)

def poll_connections(epoll, conns_write, conns_read, q_redirect, q_robots,
        q_docs, terminate):
    urls_got = 0
    DELAY_NO_EVENTS = 0.1
    DELAY_DOC_QUEUE_FULL = 0.25
    #passes = 0
    while not terminate[0]:
        #do not download more documents when the documents queue is full
        if config.MAX_DOC_QUEUE < len(q_docs):
            sleep(DELAY_DOC_QUEUE_FULL)
            continue
        #poll connections, send & receive data
        events = epoll.poll(timeout=0)
        if not events:
            sleep(DELAY_NO_EVENTS)
            continue
        for conn_id, event in events:
            #handle connections ready for writing data
            if event & select.EPOLLOUT:
                #get conn object from the write queue
                conn = conns_write.pop_(conn_id, None)
                if conn:
                    #call write handle
                    try:
                        conn.handle_write()
                    except http.ConnectionException as e:
                        if conn.is_robot():
                            q_robots.add_((conn.get_host(), util.MSG_ROBOT_FAIL))
                        continue
                    #switch the connection to reading
                    conns_read.set_(conn_id, conn)
                    epoll.modify(conn_id, select.EPOLLIN)
            #handle connections ready for reading data
            if event & (select.EPOLLIN | select.EPOLLPRI):
                conn = conns_read.get(conn_id, None)
                if conn:
                    #call read handle
                    read_result = ()
                    try:
                        read_result = conn.handle_read()
                    except http.RedirectException as e:
                        #handle the redirection
                        if conn.is_robot():
                            #redirecting robots.txt is ugly/intricate
                            #q_robots.add_((conn.get_host(), util.MSG_ROBOT_FAIL))
                            #our fault but who cares => allow all paths
                            q_robots.add_((conn.get_host(), util.MSG_ROBOT_EMPTY))
                            logging.warning('Ignoring redirected robots.txt at %s' %
                                conn.get_host().encode('utf-8'))
                        else:
                            #write out the redirection url to get
                            # it back along with the host's IP address
                            q_redirect.add_((conn.get_host(), e.get_redirect_url()))
                    except http.IncorrectResponseError as e:
                        if conn.is_robot():
                            if e.is_remote_fault():
                                #not our fault => allow all paths
                                q_robots.add_((conn.get_host(), util.MSG_ROBOT_EMPTY))
                            else:
                                q_robots.add_((conn.get_host(), util.MSG_ROBOT_FAIL))
                    except (http.HTTPException, http.ConnectionException, 
                            util.CheckFailException) as e:
                        if conn.is_robot():
                            q_robots.add_((conn.get_host(), util.MSG_ROBOT_FAIL))
                    #handle the read result
                    try:
                        keep_conn, read_doc_data = read_result
                    except ValueError:
                        keep_conn, read_doc_data = (False, ())
                    #remove the connection from the read queue
                    if not keep_conn:
                        try:
                            conns_read.remove_(conn_id)
                        except KeyError:
                            logging.debug('Failed to remove a connection from conns_read')
                    #put the result document into documents queue
                    try:
                        header, body = read_doc_data
                    except ValueError:
                        continue
                    if conn.is_robot():
                        q_robots.add_((conn.get_host(), body))
                    else:
                        q_docs.add((conn, header, body))
                    urls_got += 1

def check_conns_alive(conns_write, conns_read, q_robots, terminate):
    #stop inactive connections    
    CHECK_PERIOD = 40
    while not terminate[0]:
        deleted = 0
        conns_write.acquire()
        conns_read.acquire()
        try:
            for conn_list in (conns_write, conns_read):
                for conn_id in conn_list.keys():
                    conn = conn_list[conn_id]
                    if not conn.check_last_action():
                        if conn.is_robot():
                            q_robots.add_((conn.get_host(), util.MSG_ROBOT_FAIL))
                        del(conn_list[conn_id])
                        deleted += 1
            if deleted:
                logging.debug('inactivity timeout: %d conns closed' % deleted)
        finally:
            conns_write.release()
            conns_read.release()
        sleep(CHECK_PERIOD)

def send_to_evaluate(q_redirect, q_robots, robot_bytes, terminate):
    NO_ACTION_DELAY = 10
    while not terminate[0]:
        idle = True
        #robots
        if q_robots:
            while q_robots:
                host, robot_lines = q_robots.pop_()
                robot_bytes[0] += len(robot_lines)
                send_robot_lines(host, robot_lines)
            idle = False
        #redirection urls
        if q_redirect:
            q_redirect.acquire()
            try:
                send_redirect_urls(q_redirect)
                q_redirect.clear()
            finally:
                q_redirect.release()
            idle = False
        if idle:
            sleep(NO_ACTION_DELAY)
        sleep(2)

def send_to_process(proc_id, q_docs, doc_bytes, terminate):
    NO_DOCS_DELAY = 0.5
    #write to document processor's pipe until terminate
    doc_proc_pipe = open('%s-%d' % (util.PIPE_PATH, proc_id), 'w')
    try:
        while not terminate[0]:
            #get document data
            try:
                conn, header, body = q_docs.pop()
            except KeyError:
                sleep(NO_DOCS_DELAY)
                continue
            doc_bytes[proc_id] += len(header) + len(body)
            #send data to the document processor
            data = '%s\n%s\n%s\n%s\n%s\n%s\n\n%s' % (
                conn.get_ip().encode('utf-8'), conn.get_connect_time(), 
                conn.get_host().encode('utf-8'), conn.get_path().encode('utf-8'), 
                conn.get_url().encode('utf-8'), header, body)
            try:
                util.send_data(doc_proc_pipe, data, in_unicode=False)
            except IOError:
                #the doc processor has been terminated
                logging.warning('DOC-PROC %d TERMINATED' % proc_id)
                break
    finally:
        try:
            util.send_data(doc_proc_pipe, util.TERM_ORDER, flush=True)
        except IOError:
            pass
        logging.info('Writing to doc-proc #%d terminated' % proc_id)
        doc_proc_pipe.close()

def print_info(url_queue, conns_write, conns_read, q_redirect, q_robots, 
        q_docs, doc_bytes, robot_bytes, terminate):
    if not config.INFO_PERIOD:
        return
    period_start = util.now()
    while not terminate[0]:
        if 1 > util.seconds_left(period_start, config.INFO_PERIOD):
            period_start = util.now()
            total_bytes = sum(doc_bytes.values()) + robot_bytes[0]
            logging.info('QUEUE: %d | MB got: %d | connsW: %d, connsR: %d | redirQ: %d, '
                    'robotQ: %d, docQ: %d' %
                (len(url_queue), total_bytes / 1024**2, len(conns_write), 
                len(conns_read), len(q_redirect), len(q_robots), len(q_docs)))
        sleep(30)

epoll = select.epoll() #connections poller
conns_write = util.SafeDict() #open connections to write to
conns_read = util.SafeDict() #open connections to read from
url_queue = util.SafeDeque() #urls waiting to be downloaded
q_robots = util.SafeSet() #robots.txt data waiting to be sent
q_redirect = util.SafeSet() #redirect urls waiting to be sent
q_docs = set() #documents waiting to be sent
robot_bytes = [0] #robot.txt bytes downloaded
doc_bytes = {} #document bytes downloaded by doc processor fileno
terminate = [False] #termination flag
try:
    #obtain IDs of the external document processors
    doc_proc_ids = map(int, util.recv_data(sys.stdin).split('\t'))
    #thread to receive & process new urls
    t_urls = Thread(target=receive_data, name='receive_data',
        args=(url_queue, q_docs, terminate))
    t_urls.start()
    #thread creating new connections
    t_conns = Thread(target=create_connections, name='create_connections',
        args=(url_queue, epoll, q_robots, conns_write, conns_read, 
        q_docs, terminate))
    t_conns.start()
    #thread polling the connections
    t_poll = Thread(target=poll_connections, name='poll_connections',
        args=(epoll, conns_write, conns_read, q_redirect, q_robots, q_docs,
        terminate))
    t_poll.start()
    #thread to close inactive connections
    t_alive = Thread(target=check_conns_alive, name='check_conns_alive',
        args=(conns_write, conns_read, q_robots, terminate))
    t_alive.start()
    #thread sending data to evaluate
    t_send_eval = Thread(target=send_to_evaluate, name='send_eval',
        args=(q_redirect, q_robots, robot_bytes, terminate))
    t_send_eval.start()
    #threads sending documents to process
    t_send_proc = {}
    for proc_id in doc_proc_ids:
        doc_bytes[proc_id] = 0
        t_send_proc[proc_id] = Thread(target=send_to_process,
            name='send_proc', args=(proc_id, q_docs, doc_bytes, terminate))
        t_send_proc[proc_id].start()
    #thread to print debug information periodically
    t_info = Thread(target=print_info, name='print_info',
        args=(url_queue, conns_write, conns_read, q_redirect, q_robots,
        q_docs, doc_bytes, robot_bytes, terminate))
    t_info.start()
    #wait for the threads to finish
    t_urls.join()
    t_conns.join()
    t_poll.join()
    t_alive.join()
    t_send_eval.join()
    for proc_id in doc_proc_ids:
        t_send_proc[proc_id].join()
    t_info.join()
except Exception as e:
    logging.critical(e)
finally:
    logging.info('Total MB downloaded: %.2f' % 
        ((sum(doc_bytes.values()) + robot_bytes[0]) / 1024.0**2))
    epoll.close()
    logging.debug('Crawler stopped')
