#===============================================================================
#   SpiderLing -- A web spider/crawler for linguistic purpose.
#   Copyright (C) 2011-2015  Vit Suchomel
#   http://corpus.tools/wiki/SpiderLing
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================

'''
input: original encoding
arc: original encoding
prevert: utf-8
links: utf-8
store_arc: always, raw data
store_prevertical: successfully decoded, text extraction ok, language check ok
get_links: html not empty, successful decoding not required
'''

import sys, os, re, codecs
import gzip
import cStringIO as StringIO
import subprocess

import util
from util import config
if config.UNIGRAM_MODELS:
    from util import trigrams1 as trigrams #unigram hack for CJK
else:
    from util import trigrams
from chared.detector import EncodingDetector
from justext import core as justext
from util.http import url_split, url_join_rel_norm

BAD_FILE_EXTENSIONS = ['ai','aiff','asf','avi','bmp','bz2','c','css','deb',
                     'djvu','dvi','eps','gif','gz','h','ico','jpg',
                     'jpeg','mid','mng','mov','mp3','mpg','ods','ogg','pas',
                     'png','ppt','psd','qt','ra','ram','rm','rpm',
                     'rtf','sdd','sdw','sit','swf','sxc','sxi','sxw','tex',
                     'tgz','tif','wav','wmv','xcf','xls','xml','zip']
BAD_FILE_EXTENSIONS_RE = re.compile('\.(?:%s)$' % '|'.join(BAD_FILE_EXTENSIONS), re.I)
BIN_FILE_EXTENSIONS_RE = re.compile('\.(?:doc|docx|pdf|ps)$', re.I)
META_ENC1_RE = re.compile('''<meta\s+http-equiv=['"]?content-type['"]?\s+content=['"]?[^'"]*charset=([^'" >/]+)''', re.I)
META_ENC2_RE = re.compile('''<meta\s+content=['"]?[^'"]*charset=([^'" >/]+)['"]?\s+http-equiv=['"]?content-type['"]?''', re.I)
META_ENC3_RE = re.compile('''<meta\s+http-equiv=['"]?charset['"]?\s+content=['"]?([^'" >/]+)''', re.I)
META_ENC4_RE = re.compile('''<meta\s+content=['"]?([^'" >/]+)['"]?\s+http-equiv=['"]?charset['"]?''', re.I)
META_ENC5_RE = re.compile('''<meta\s+charset=['"]?([^'" >/]+)''', re.I)

URL_RE = re.compile('https?://.+\..+', re.I)
URL_SPLIT_CHARS = '/&'
FILE_EXT_RE = re.compile(u'\.([^.]*)$')
CONVERSION_COMMANDS = {
    'pdf': ['pdftotext', '-nopgbrk', '-', '-'],
    'ps': ['ps2ascii', '-'],
    'doc': ['antiword', '-mUTF-8', '-'],
    'docx': ['util/docx2txt.pl', '-', '-'],
}
CONVERSION_TIMEOUT = ['timeout', '-k', '10', '30'] #send TERM/KILL to make sure the command ends

STDIN_WAIT_TIME = 0.1
DOC_NOT_ACCEPTED = -1
DOC_EMPTY = 0
MIN_WORDLIST_HITS = 3
HTML_BASE_TAG_RE = re.compile('<base [^>]*href="([^"]+)/?"', re.UNICODE)

import logging
logging.basicConfig(filename='%s.log.process-%d' % (config.PREFIX, os.getpid()), filemode='a',
    level=config.LOG_LEVEL, format='%(asctime)-15s PROCESS %(message)s')

def create_lang_model(language):
    sample_path = os.path.join('util', 'lang_samples', language)
    with codecs.open(sample_path, mode='r', encoding='utf-8') as fp:
        lang_sample = fp.read()
    lang_model = trigrams.Trigram()
    lang_model.parseLines([lang_sample])
    return lang_model

def get_lang_difference(lang_model, txt_data):
    data_trigrams = trigrams.Trigram()
    data_trigrams.parseLines([txt_data])
    lang_difference = lang_model - data_trigrams
    return lang_difference

def format_date(timestamp):
    if 12 == len(timestamp):
        return '20%s-%s-%s %s:%s' % (timestamp[0:2], timestamp[2:4], timestamp[4:6],
            timestamp[6:8], timestamp[8:10])
    logging.warning('Invalid crawl_date value: %s' % timestamp)
    return timestamp

def format_doc_len(doc_len):
    if 1000 > doc_len:
        return '0-1k'
    if 10000 > doc_len:
        return '1k-10k'
    if 100000 > doc_len:
        return '10k-100k'
    if 1000000 > doc_len:
        return '100k-1M'
    return '1M-oo'

def convert_to_txt(command, data):
    dataIO = StringIO.StringIO()
    dataIO.write(data)
    dataIO.seek(0)
    try:
        process = subprocess.Popen(CONVERSION_TIMEOUT + command,
            stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        process_output, process_errors = process.communicate(dataIO.read())
    except Exception as e:
        logging.debug('convert_to_txt exception: %s' % e)
        return ''
    finally:
        dataIO.close()
    if process_errors:
        pass #error output is ignored since some tools use it for general reporting
    return process_output

def html_decode(html_string, chared_model):
    #find encoding specified in the header (copied from jusText)
    meta_encoding = None
    for re_meta in (META_ENC1_RE, META_ENC2_RE, META_ENC3_RE, META_ENC4_RE, META_ENC5_RE):
        m = re_meta.search(html_string)
        if m:
            try:
                meta_encoding = codecs.lookup(m.group(1)).name
            except LookupError:
                meta_encoding = None
                logging.debug('Invalid meta encoding: %s' % m.group(1))
            else:
                break
    #detect encoding using chared
    chared_encodings = chared_model.classify(html_string)
    chared_encoding_0 = chared_encodings[0] if chared_encodings else None
    #meta not detected => use chared
    if not meta_encoding:
        for chared_encoding in chared_encodings:
            try:
                return (unicode(html_string, chared_encoding, errors='strict'), meta_encoding,
                    chared_encoding)
            except:
                pass
    #meta + chared agreement => use meta
    elif meta_encoding in chared_encodings:
        try:
            return (unicode(html_string, meta_encoding, errors='strict'), meta_encoding,
                meta_encoding)
        except:
            pass
    #no agreement and no vote for utf-8 => use meta, if that fails, use chared
    elif 'utf-8' != meta_encoding and 'utf-8' not in chared_encodings:
        try:
            return (unicode(html_string, meta_encoding, errors='strict'), meta_encoding,
                chared_encoding_0)
        except:
            pass
        for chared_encoding in chared_encodings:
            try:
                return (unicode(html_string, chared_encoding, errors='strict'), meta_encoding,
                    chared_encoding)
            except:
                pass
    #try utf-8 if everything fails
    return (unicode(html_string, 'utf-8', errors='strict'), meta_encoding, chared_encoding_0)

def txt_decode(txt_string, chared_model):
    #detect encoding using chared
    chared_encodings = chared_model.classify(txt_string)
    for chared_encoding in chared_encodings:
        try:
            return (unicode(txt_string, chared_encoding, errors='strict'), None, chared_encoding)
        except:
            pass
    #try utf-8 if everything fails
    return unicode(txt_string, 'utf-8', errors='strict')

def extract_paragraphs(input_name, input_data, chared_model, force_encoding=None):
    meta_encoding = chared_encoding = None
    if 'html' in input_data[:100]:
        fext = 'html'
    else:
        m = FILE_EXT_RE.search(input_name)
        fext = m.group(1) if m else 'html'
    #!TODO use file header instead of file name to recognise the type
    if fext == 'txt' or fext in CONVERSION_COMMANDS:
        #extract paragraphs from non HTML data using a convertor and Justext
        if fext == 'txt':
            txt_data = input_data
        else:
            txt_data = convert_to_txt(CONVERSION_COMMANDS[fext], input_data)
        if force_encoding:
            unicode_txt_data = unicode(txt_data, force_encoding, errors='replace')
        else:
            try:
                unicode_txt_data, meta_encoding, chared_encoding = txt_decode(txt_data,
                    chared_model)
            except:
                logging.debug('Could not txt_decode document %s' % input_name)
                return
        paragraphs = []
        if unicode_txt_data:
            for par_text in unicode_txt_data.split(u'\n\n'):
                paragraphs.append({
                    'text': par_text,
                    'word_count': len(par_text.split()),
                    'linked_char_count': 0,
                    'dom_path': 'p',
                })
        html_root = None
        html_text = ''
    else:
        #extract paragraphs from HTML data using Justext
        if force_encoding:
            unicode_input_data = unicode(input_data, force_encoding, errors='replace')
        else:
            try:
                unicode_input_data, meta_encoding, chared_encoding = html_decode(input_data,
                    chared_model)
            except:
                logging.debug('Could not html_decode document %s' % input_name)
                return
        if unicode_input_data:
            html_text = unicode_input_data
            utf8_input_data = unicode_input_data.encode('utf-8')
            html_root = justext.preprocess(html_text=utf8_input_data, encoding='utf-8')
            paragraphs = justext.make_paragraphs(html_root)
        else:
            paragraphs, html_root, html_text = [], None, ''
    return (paragraphs, html_root, html_text, meta_encoding, chared_encoding)

"Extract text from HTML pages using Justext or from other files using a convertor."
def any2txt(input_name, input_data, lang_diff_threshold_doc, lang_diff_threshold_par,
        chared_models, lang_models, justext_stoplists, force_encoding=None):
    #extract paragraphs of plaintext from the input document
    #try for all languages until the difference is below the difference threshold
    html_root = None
    html_text = meta_encoding = chared_encoding = ''
    min_doc_lang_diff = 1.0
    for lang, lang_model in lang_models.iteritems():
        paragraph_data = extract_paragraphs(input_name, input_data, chared_models[lang],
            force_encoding)
        if paragraph_data:
            paragraphs, html_root, html_text, meta_encoding, chared_encoding = paragraph_data
            doc_text = u'\n'.join([p['text'] for p in paragraphs])
            if doc_text:
                doc_lang_diff = get_lang_difference(lang_model, doc_text)
                if doc_lang_diff < min_doc_lang_diff:
                    doc_lang = lang
                    min_doc_lang_diff = doc_lang_diff
    if min_doc_lang_diff > lang_diff_threshold_doc:
        return #different from all recognised languages

    #use Justext to classify paragraphs -- quite permissive setting
    justext.classify_paragraphs(
        paragraphs=paragraphs,
        stoplist=justext_stoplists[doc_lang],
        length_low=30, #default=70 #< length_low => bad or short
        length_high=100, #default=200 #> length_high => good
        stopwords_low=0.2, #default=0.30 #>= stopwords_low => neargood
        stopwords_high=0.3, #default=0.32 #>= stopwords_high => good or neargood
        max_link_density=0.4, #default=0.2 #link_density > max_link_density => bad
    )
    justext.revise_paragraph_classification(
        paragraphs=paragraphs,
        max_heading_distance=400, #default=200 #Short/near-good heads in the distance [chars] before a good par => good
    )

    #return plaintext and prevertical with <p> tags
    plaintext = []
    prevertical = []
    for p in paragraphs:
        if p['cfclass'] in ('good', 'neargood'): #'good', 'neargood', 'short', 'bad'
            p_text = justext.html_escape(p['text'])
            lang_difference = get_lang_difference(lang_models[doc_lang], p_text)
            if lang_difference <= lang_diff_threshold_par:
                plaintext.append(p_text)
                heading = u' heading="1"' if p['heading'] else u''
                prevertical.append(u'<p%s langdiff="%.2f">%s</p>' %
                    (heading, lang_difference, p_text))
    if plaintext:
        return (u'\n'.join(plaintext), u'\n'.join(prevertical),
            html_root, html_text, doc_lang, min_doc_lang_diff, meta_encoding, chared_encoding)

def extract_links(html_root, html, url, host, domain_blacklist, unprocessed_urls_fp):
    if html_root is None:
        return set()
    doc_links = []
    #check base tag
    m = HTML_BASE_TAG_RE.search(html)
    base_url = u'%s/' % m.group(1) if m else url
    #extract links
    for a_tag in html_root.iterfind('.//a[@href]'):
        new_url = a_tag.get('href')
        if new_url:
            #resolve relative paths, join with base url, normalise (no trailing slash)
            try:
                doc_links.append(url_join_rel_norm(base_url, new_url).strip())
            except ValueError:
                logging.debug('extract_links\turl_join_rel_norm fail\t%s' % url)
                continue
    #filter out bad links
    result = set()
    all_links = set()
    for url in doc_links:
        #omit local duplicates
        if url in all_links:
            continue
        all_links.add(url)
        #omit invalid or too long urls
        if not URL_RE.match(url) or len(url) < 12 or len(url) > 400 or url.count('/') > 30 or \
                url.count('.') > 30:
            continue
        #omit urls consisting of many same parts
        for split_char in URL_SPLIT_CHARS:
            split_parts = url.split(split_char)
            if len(set(split_parts)) + 10 < len(split_parts):
                continue
        #split to parts
        try:
            url_schema, url_host, url_path = url_split(url)
        except ValueError:
            logging.debug('extract_links\turl_split fail\t%s' % url)
            continue
        #omit empty or unwanted domains:
        if not url_host or url_host in domain_blacklist:
            continue
        #omit unwanted file extensions
        if BAD_FILE_EXTENSIONS_RE.search(url_path):
            continue
        #do not process binary files (workaround) #!TODO
        if BIN_FILE_EXTENSIONS_RE.search(url_path):
            unprocessed_urls_fp.write('%s\n' % url)
            continue
        #check top level domain
        if config.ALLOWED_TLDS_RE.search(url_host):
            result.add((url_schema, url_host, url_path))
            continue
        #check allowed site domains
        for dom in config.ALLOWED_EXTRA_DOMAINS:
            if dom in url:
                result.add((url_schema, url_host, url_path))
                continue
    return result

#read this processor's ID
doc_processor_id = int(util.recv_data(sys.stdin))
logging.info('Document processor #%d started' % doc_processor_id)

#init
raw_data_count = 0
lang_models = {}
chared_models = {}
justext_stoplists = {}
for lang in config.LANGUAGES:
    lang_models[lang] = create_lang_model(lang)
    chared_lang_model_path = os.path.join(config.CHARED_PATH, 'models', lang)
    chared_models[lang] = EncodingDetector.load(chared_lang_model_path)
    justext_stoplists[lang] = justext.get_stoplist(lang)

#domain blacklist
domain_blacklist = set()
if config.DOMAIN_BLACKLIST_FILE:
    domain_blacklist.update(open(config.DOMAIN_BLACKLIST_FILE).read().strip().split('\n'))
#open files
arc_file = '%s.%d.arc.gz' % (config.PREFIX, doc_processor_id)
arc_fp = gzip.open(arc_file, mode='w', compresslevel=9)
prevert_file = '%s.%d.prevert_d' % (config.PREFIX, doc_processor_id)
prevert_fp = codecs.open(prevert_file, mode='w', encoding='utf-8')
unprocessed_urls_file = '%s.%d.unproc_urls' % (config.PREFIX, doc_processor_id)
unprocessed_urls_fp = codecs.open(unprocessed_urls_file, mode='w', encoding='utf-8')

#read documents from input_pipe until terminate
input_pipe = open('%s-%d' % (util.PIPE_PATH, doc_processor_id), 'r')
try:
    while True:
        input_data = util.recv_data(input_pipe, in_unicode=False)
        if not input_data:
            continue
        #detect termination order
        if util.TERM_ORDER == input_data:
            logging.info('TERMINATE')
            break
        #process the next document
        raw_data_count += 1
        doc_id = '%d-%d' % (doc_processor_id, raw_data_count) #unique doc ID
        txt = prevert = html = ''
        html_root = meta_encoding = chared_encoding = None
        #extract document meta-attributes and data
        ip, conn_time, host, path, url, doc_data = input_data.split('\n', 5)
        host = unicode(host, 'utf-8')
        path = unicode(path, 'utf-8')
        url = unicode(url, 'utf-8')
        header, body = doc_data.split('\n\n', 1)
        #extract clean text and prevertical
        try:
            extracted_data = any2txt(path, body, config.LANG_DIFF_THRESHOLD_DOC,
                config.LANG_DIFF_THRESHOLD_PAR, chared_models, lang_models,
                justext_stoplists, config.FORCE_ENCODING)
        except Exception as e:
            try:
                logging.debug('Text conversion exception: %s' % e)
            except:
                logging.debug('Text conversion exception')
        if extracted_data is not None and extracted_data[4] in config.LANGUAGES_ACCEPT:
            txt, prevert, html_root, html, lang, lang_diff, meta_enc, chared_enc = extracted_data
        #send document's attributes and links
        links = extract_links(html_root, html, url, host, domain_blacklist, unprocessed_urls_fp)
        links_txt = '\n'.join(map(lambda x: '\t'.join(x), links))
        util.send_data(sys.stdout, u'%s\n%s\n%d\n%d\n%d\n%d\n%s' %
            (doc_id, host, len(html), hash(html), len(txt), hash(txt), links_txt))
        #store in arc format
        arc_fp.write('%s %s %s text/html %d %s\n%s\n' %
            (url.encode('utf-8'), ip, conn_time, len(doc_data), doc_id, doc_data))
        #store in prevertical format
        if txt:
            try:
                prevert_fp.write(u'<doc id="%s" length="%s" crawl_date="%s" lang="%s" '
                    'lang_diff="%.2f" ip="%s" url="%s" enc_meta="%s" enc_chared="%s">\n%s\n</doc>\n' %
                    (doc_id, format_doc_len(len(txt)), format_date(conn_time), lang, lang_diff,
                    ip, util.url_escape(url), meta_enc, chared_enc, prevert))
            except UnicodeDecodeError:
                logging.debug('UnicodeDecodeError while storing a prevert of %s' % url)
        else:
            logging.debug('No clean text for %s%s' % (host, path))
finally:
    try:
        #send out a notice this document processor is stopped
        util.send_data(sys.stdout, util.TERM_RESPONSE, flush=True)
    finally:
        prevert_fp.close()
        arc_fp.close()
        unprocessed_urls_fp.close()
        input_pipe.close()
        logging.info('Document processor #%d stopped' % doc_processor_id)
