#!/usr/bin/pypy

#===============================================================================
#   SpiderLing -- A web spider/crawler for linguistic purpose.
#   Copyright (C) 2011-2015  Vit Suchomel
#   http://corpus.tools/wiki/SpiderLing
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================

USAGE = """
Reprocess arc file using document processors and respective configuration.
Useful for re-processing previously crawled raw arc data.
zcat spiderling.*.arc.gz | pypy reprocess.py [DOC_PROCESSOR_COUNT]
"""

import sys, os, codecs
from time import sleep

import util
from util import config
from util.http import url_split

import logging
logging.basicConfig(filename='%s.log.reprocess' % config.PREFIX, filemode='a',
    level=config.LOG_LEVEL, format='%(asctime)-15s %(message)s')

class DocProcessorWithoutDomains(util.SubProcess):
    PROCESS_SCRIPT = 'process.py'

    def __init__(self, id_, html_hashes, txt_hashes, arc_queue):
        self._id = id_ + 100
        super(DocProcessorWithoutDomains, self).__init__(name='doc-%d' % self._id)
        self.html_hashes = html_hashes
        self.txt_hashes = txt_hashes
        self.arc_queue = arc_queue
        self._html_count = self._html_size = self._txt_count = self._txt_size = 0
        self._pipe_path = '%s-%d' % (util.PIPE_PATH, self._id)
        os.mkfifo(self._pipe_path)

    def read(self):
        duplicates_path = '%s.%d.duplicates' % (config.PREFIX, self._id)
        with codecs.open(duplicates_path, mode='w', encoding='utf-8') as duplicates_fp:
            while True: #read until util.TERM_RESPONSE
                data = util.recv_data(self._subproc.stdout)
                if util.TERM_RESPONSE == data:
                    break
                if data:
                    doc_id, host, html_length, html_hash, txt_length, txt_hash, new_urls_msg \
                        = data.split('\n', 6)
                    html_length = int(html_length)
                    txt_length = int(txt_length)
                    self._html_count += 1
                    self._html_size += html_length
                    try:
                        self.html_hashes.add_unique(int(html_hash))
                    except util.ItemNotUniqueException:
                        if txt_length:
                            duplicates_fp.write('%s\n' % doc_id)
                        continue
                    if txt_length:
                        try:
                            self.txt_hashes.add_unique(int(txt_hash))
                        except util.ItemNotUniqueException:
                            duplicates_fp.write('%s\n' % doc_id)
                            continue
                        self._txt_count += 1
                        self._txt_size += txt_length
                else:
                    sleep(self.IO_WAIT_DELAY)

    def write(self):
        util.send_data(self._subproc.stdin, str(self._id), flush=True) #doc processor's ID
        with open(self._pipe_path, 'w') as subproc_pipe_fp:
            while True: #write until util.TERM_ORDER
                if self._stopped:
                    data = util.TERM_ORDER
                else:
                    self.arc_queue.acquire()
                    try:
                        data = self.arc_queue.popleft()
                    except IndexError:
                        sleep(self.IO_WAIT_DELAY)
                        continue
                    finally:
                        self.arc_queue.release()
                try:
                    util.send_data(subproc_pipe_fp, data, flush=True, in_unicode=False)
                except IOError as e:
                    logging.error(e)
                if self._stopped:
                    self._subproc.stdin.flush()
                    break
        try:
            os.unlink(self._pipe_path)
        except IOError as e:
            logging.error(e)

    def get_counts(self):
        return (self._html_count, self._html_size, self._txt_count, self._txt_size)

if '-h' in sys.argv[1:] or '--help' in sys.argv[1:]:
    sys.stderr.write(USAGE)
    sys.exit(1)
if len(sys.argv) > 1:
    doc_proc_count = int(sys.argv[1])
else:
    doc_proc_count = config.DOC_PROCESSOR_COUNT
stdin = sys.stdin

logging.info('Reprocess starts with %d document processors' % doc_proc_count)
html_hashes, txt_hashes = util.SafeUniqueSet(), util.SafeUniqueSet()
arc_queue = util.SafeDeque()
util.reset_dir(config.PIPE_DIR)
doc_processors = {}
for p_id in range(doc_proc_count):
    doc_processors[p_id] = DocProcessorWithoutDomains(p_id, html_hashes, txt_hashes, arc_queue)
sleep(3)
for p_id in range(doc_proc_count):
    doc_processors[p_id].IO_threads_start()
logging.info('Sending arc data from stdin to document processors')
arc_count = 0
while True:
    try:
        url, ip, conn_time, mime_type, chunk_len, doc_id = stdin.readline().rstrip().split(' ')
    except ValueError:
        break
    size_to_read = int(chunk_len)
    doc_data = stdin.read(size_to_read)
    stdin.read(1)
    if len(doc_data) < size_to_read:
        break
    arc_count += 1
    url = unicode(url, 'utf-8')
    try:
        scheme, host, path = url_split(url)
    except ValueError:
        host, path = '[unknown host]', '[unknown path]'
    while len(arc_queue) > config.MAX_DOC_QUEUE:
        sleep(1)
    arc_queue.acquire()
    arc_queue.append('\n'.join((ip, conn_time, host.encode('utf-8'), path.encode('utf-8'),
        url.encode('utf-8'), doc_data)))
    arc_queue.release()
logging.info('Reading %d arc records finished' % arc_count)
while len(arc_queue) > 1:
    sleep(1)
sleep(3)
html_count = html_size = txt_count = txt_size = 0
for p_id in range(doc_proc_count):
    doc_processors[p_id].stop()
    doc_processors[p_id].IO_threads_join()
    doc_processors[p_id].join_subprocess()
    p_html_count, p_html_size, p_txt_count, p_txt_size = doc_processors[p_id].get_counts()
    html_count += p_html_count
    html_size += p_html_size
    txt_count += p_txt_count
    txt_size += p_txt_size
logging.info('All finished, html recv %d (%.1f MB), txt recv %d (%.1f MB)' %
    (html_count, float(html_size) / 1024**2, txt_count, float(txt_size) / 1024**2))
