#!/usr/bin/pypy

#===============================================================================
#   SpiderLing -- A web spider/crawler for linguistic purpose.
#   Copyright (C) 2011-2015  Vit Suchomel
#   http://corpus.tools/wiki/SpiderLing
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================

USAGE = """
A web spider/crawler for linguistic purpose. See README.
pypy spiderling.py SEED_URLS [SAVEPOINT_TIMESTAMP]
"""

import sys, os, codecs
import signal
from time import sleep
from threading import Thread

import util
from util import config
from util.domain import DomainQueue
from util.http import url_split

F_DOMS = '%s.state-%s-domains'
F_DOM_DISTS = '%s.state-%s-domain-distances'
F_HASHES_HTML = '%s.state-%s-hashes-html'
F_HASHES_TXT = '%s.state-%s-hashes-txt'

import logging
logging.basicConfig(filename='%s.log.eval' % config.PREFIX, filemode='a',
    level=config.LOG_LEVEL, format='%(asctime)-15s %(message)s')

class Crawler(util.SubProcess):
    PROCESS_SCRIPT = 'crawl.py'
    MIN_URL_SEND = 40
    MAX_URL_SEND = 80

    def __init__(self, name, domains, domain_distances):
        super(Crawler, self).__init__(name, logging)
        self._html_sent_count = 0
        self.domains = domains
        self.domain_distances = domain_distances
        self._doms_dead = util.SafeSet()
        Thread(target=self.check_dead_domains_loop, name='dom-dead').start()

    def check_dead_domains_loop(self):
        PERIOD = 240
        period_start = util.now()
        while not self._stopped:
            if 1 > util.seconds_left(period_start, PERIOD):
                self._doms_dead.acquire()
                self._doms_dead.update(domains.get_doms_dead())
                self._doms_dead.release()
                period_start = util.now()
            sleep(60)

    def send_doc_processors_ids(self, doc_proc_ids):
        #tell the crawler the document processors' IDs
        doc_proc_ids = '\t'.join(map(str, doc_proc_ids))
        util.send_data(self._subproc.stdin, doc_proc_ids, flush=True)

    def get_send_data(self):
        if self._doms_dead:
            self._doms_dead.acquire()
            doms_dead_str = '\t'.join(self._doms_dead) 
            self._doms_dead.clear()
            self._doms_dead.release()
            return u'%s\t%s' % (util.MSG_DEAD_DOMS, doms_dead_str)
        else:
            urls = self.domains.get_good_urls()
            if urls:
                self._html_sent_count += len(urls)
                return u'\n'.join(map(u'\t'.join, urls))

    def process_recv_data(self, data):
        message_type, message = data.split('\n', 1)
        #robot lines
        if message_type == util.MSG_ROBOT_LINES:
            try:
                host, robot_lines = message.split('\n', 1)
            except Exception:
                logging.error('Bad robot message data => the whole domain will be dropped')
            try:
                domain = self.domains[host]
            except KeyError:
                logging.warning('Non existing domain %s (robots)' % host)
            else:
                if domain:
                    domain.set_robot_lines(robot_lines)
                else: #removed domain
                    logging.warning('Removed domain %s (robots)' % host)
        #redirect urls sent back by the crawler to process
        elif util.MSG_NEW_URL == message_type:
            for url_pair in message.splitlines():
                try:
                    src_host, url = url_pair.split('\t')
                except ValueError:
                    logging.error('Bad redirect message data => a url will be skipped')
                else:
                    process_new_urls([url], src_host, self.domains, self.domain_distances)
        else:
            logging.warning('Unprocessed crawler message: %s' % message_type)

    def get_html_sent_count(self):
        return self._html_sent_count

class DocProcessor(util.SubProcess):
    PROCESS_SCRIPT = 'process.py'
    PYTHON = 'python' #process.py requires lxml which does not work with pypy

    def __init__(self, id, name, domains, domain_distances, html_hashes, txt_hashes):
        super(DocProcessor, self).__init__(name, logging)
        self._id = id
        self._html_count = 0
        self._html_size = 0
        self.html_hashes = html_hashes
        self._txt_count = 0
        self._txt_size = 0
        self.txt_hashes = txt_hashes
        self.domains = domains
        self.domain_distances = domain_distances

    def read(self):
        #keep a duplicates file opened while reading
        duplicates_path = '%s.%d.duplicates' % (config.PREFIX, self._id)
        self._duplicates_fp = codecs.open(duplicates_path, mode='w', encoding='utf-8')
        try:
            super(DocProcessor, self).read()
        finally:
            self._duplicates_fp.close()

    def write(self):
        #send doc processor's ID to the subprocess
        util.send_data(self._subproc.stdin, str(self._id), flush=True)

    def process_recv_data(self, data):
        doc_id, host, html_length, html_hash, txt_length, \
            txt_hash, new_urls_msg = data.split('\n', 6)
        html_length = int(html_length)
        txt_length = int(txt_length)
        self._html_count += 1
        self._html_size += html_length
        #update domain
        try:
            domain = self.domains[host]
        except KeyError:
            logging.warning('Non existing domain %s (html)' % host)
            return
        if domain:
            domain.add_size_downloaded(html_length)
        #add new links
        new_urls = new_urls_msg.splitlines()
        process_new_urls(new_urls, host, self.domains, self.domain_distances)
        #html
        try:
            self.html_hashes.add_unique(int(html_hash))
        except util.ItemNotUniqueException:
            #not unique => store duplicate doc ID
            if txt_length:
                self._duplicates_fp.write('%s\n' % doc_id)
            return
        #txt
        if txt_length:
            try:
                self.txt_hashes.add_unique(int(txt_hash))
            except util.ItemNotUniqueException:
                #not unique => store duplicate doc ID
                self._duplicates_fp.write('%s\n' % doc_id)
                return
            self._txt_count += 1
            self._txt_size += txt_length
            #update domain
            if domain:
                domain.add_size_cleaned(txt_length)

    def get_html_count(self):
        return self._html_count
    def get_html_size(self):
        return self._html_size
    def get_txt_count(self):
        return self._txt_count
    def get_txt_size(self):
        return self._txt_size

def process_new_urls(urls, src_host, domains, domain_distances, print_progress=False):
    i = 0
    domain_distances.acquire()
    for url in urls:
        #split url parts
        try:
            schema, host, path = url.rstrip().split('\t')
        except ValueError:
            #get url parts
            try:
                schema, host, path = url_split(url)
            except ValueError:
                continue
        domain_distance = domain_distances.get(host, -1)
        src_distance = domain_distances.get(src_host, -1)
        if domain_distance < 0 or src_distance >= 0 and domain_distance > src_distance + 1:
            domain_distances[host] = src_distance + 1
        #order domains to create/update the corresponding domain
        domains.add(host, path, schema)
        i += 1
        if print_progress and 0 == i % 1000000:
            logging.debug('%d URLs loaded' % i)
    domain_distances.release()
    if print_progress:
        logging.info('%d URLs loaded' % i)

def print_info_loop(domains, crawler, doc_processors, terminate):
    if not config.INFO_PERIOD:
        return
    period_start = util.now()
    while not terminate[0]:
        if 1 > util.seconds_left(period_start, config.INFO_PERIOD):
            period_start = util.now()
            print_info(domains, crawler, doc_processors)
        sleep(30)

def print_info(domains, crawler, doc_processors):
    html_sent_count = crawler.get_html_sent_count()
    html_recv_count = html_size = txt_count = txt_size = 0
    for p in doc_processors.itervalues():
        html_recv_count += p.get_html_count()
        html_size += p.get_html_size()
        txt_count += p.get_txt_count()
        txt_size += p.get_txt_size()
    doms_new, doms_ready, doms_selected, doms_bad = domains.get_stats()
    logging.info('INFO: doms new %d, ready %d, sel %d, bad %d; url sent %d; '
        'html recv %d (%.1f MB), txt recv %d (%.1f MB)' %
        (doms_new, doms_ready, doms_selected, doms_bad, html_sent_count,
        html_recv_count, html_size / 1024.0**2, txt_count, txt_size / 1024.0**2))

def save_state(domains, domain_distances, html_hashes, txt_hashes):
    state_id = util.timestamp()
    logging.info('Saving state %s' % state_id)
    try:
        logging.info('Saving domains')
        with open(F_DOMS % (config.PREFIX, state_id), 'w') as fp:
            domains.save_to_file(fp)
        logging.info('Saving domain distances')
        with codecs.open(F_DOM_DISTS % (config.PREFIX, state_id), mode='w',
                encoding='utf-8') as fp:
            domain_distances.acquire()
            for host, distance_from_seeds in domain_distances.iteritems():
                fp.write(u'%s\t%d\n' % (host, distance_from_seeds))
            domain_distances.release()
        logging.info('Saving html hashes')
        with open(F_HASHES_HTML % (config.PREFIX, state_id), 'w') as fp:
            fp.write(util.pack_ints(set(html_hashes)))
        logging.info('Saving txt hashes')
        with open(F_HASHES_TXT % (config.PREFIX, state_id), 'w') as fp:
            fp.write(util.pack_ints(txt_hashes))
    except Exception as e:
        logging.warning("Unable to save state %s." % (state_id, str(e)))
    logging.info('State saved')

def save_state_loop(domains, domain_distances, html_hashes, txt_hashes, terminate):
    if not config.STATE_PERIOD:
        return
    period_start = util.now()
    while not terminate[0]:
        if 1 > util.seconds_left(period_start, config.STATE_PERIOD):
            period_start = util.now()
            save_state(domains, domain_distances, html_hashes, txt_hashes)
        sleep(30)

def init_state(urls_oversize_fp, load_state_id=None):
    domains = DomainQueue(urls_oversize_fp)
    domain_distances = util.SafeDict()
    if load_state_id:
        logging.info('Loading domains')
        with open(F_DOMS % (config.PREFIX, load_state_id), 'r') as fp:
            domains.load_from_file(fp)
        logging.info('Loading domain distances')
        with codecs.open(F_DOM_DISTS % (config.PREFIX, load_state_id), mode='r',
                encoding='utf-8') as fp:
            domain_distances.acquire()
            for line in fp:
                host, distance_from_seeds = line.split(u'\t')
                domain_distances[host] = int(distance_from_seeds)
            domain_distances.release()
        logging.info('Loading html hashes')
        with open(F_HASHES_HTML % (config.PREFIX, load_state_id), 'r') as fp:
            html_h = util.unpack_ints(fp.read())
        logging.info('Loading txt hashes')
        with open(F_HASHES_TXT % (config.PREFIX, load_state_id), 'r') as fp:
            txt_h = util.unpack_ints(fp.read())
    else:
        html_h, txt_h = ((), ())
    return (domains, domain_distances, util.SafeUniqueSet(html_h), util.SafeUniqueSet(txt_h))

#init/load state
if len(sys.argv) < 2:
    sys.stderr.write(USAGE)
    sys.exit(1)
if len(sys.argv) > 2:
    load_state_id = sys.argv[2]
else:
    load_state_id = None
urls_oversize_fp = open('%s.urls_oversize' % config.PREFIX, 'a')
domains, domain_distances, html_hashes, txt_hashes = init_state(urls_oversize_fp, load_state_id)

#load seed URLs
if 1 < len(sys.argv):
    seed_urls = []
    with codecs.open(sys.argv[1], 'rb', 'utf-8', errors='ignore') as fp:
        for line in fp:
            seed_urls.append(line)
    process_new_urls(seed_urls, None, domains, domain_distances, print_progress=True)

logging.info('State initialized/loaded')

time_start = util.now()
terminate = [False] #termination flag
def sigterm_handler(*args):
    logging.debug('SIGTERM caught')
    terminate[0] = True
signal.signal(signal.SIGTERM, sigterm_handler)

#run crawler, document processors,...
try:
    util.reset_dir(config.PIPE_DIR)
    #subprocesses: document processors (including I/O threads)
    doc_processors = {}
    for p_id in range(config.DOC_PROCESSOR_COUNT):
        #create named pipe for communication with the subprocess
        # crawler writes to the pipe, processor reads from the pipe
        os.mkfifo('%s-%d' % (util.PIPE_PATH, p_id))
        #create a doc processor
        p = DocProcessor(p_id, 'doc-%d' % p_id, domains, domain_distances,
            html_hashes, txt_hashes)
        doc_processors[p_id] = p
    sleep(3)
    #subprocess: crawler (including I/O threads)
    crawler = Crawler('crawler', domains, domain_distances)
    #start domains checking
    domains.start()
    #thread: print information
    info_thread = Thread(target=print_info_loop, name='print_info', 
        args=(domains, crawler, doc_processors, terminate))
    #thread: preserve domains' state
    state_thread = Thread(target=save_state_loop, name='state', 
        args=(domains, domain_distances, html_hashes, txt_hashes, terminate))
    #start processes' I/O threads and other threads
    for p_id in range(config.DOC_PROCESSOR_COUNT):
        doc_processors[p_id].IO_threads_start()
    #tell the crawler the processor's ID
    crawler.send_doc_processors_ids(range(config.DOC_PROCESSOR_COUNT))
    crawler.IO_threads_start()
    info_thread.start()
    state_thread.start()
    #remove the processors' pipes after some time
    sleep(180)
    for p_id in range(config.DOC_PROCESSOR_COUNT):
        try:
            os.unlink('%s-%d' % (util.PIPE_PATH, p_id))
        except IOError as e:
            logging.error(e)
    #wait for all threads to finish, terminate subprocesses
    while not terminate[0] and (not config.MAX_RUN_TIME or util.seconds_left(
            time_start, config.MAX_RUN_TIME)) > 0:
        sleep(30)
    logging.info('All processes will be stopped')
    terminate[0] = True
    domains.terminate()
    crawler.stop()
    crawler.IO_threads_join()
    logging.debug('JOIN-crawler')
    crawler.join_subprocess()
    logging.debug('TERMINATE-crawler')
    for p_id in range(config.DOC_PROCESSOR_COUNT):
        doc_processors[p_id].stop()
        doc_processors[p_id].IO_threads_join()
        logging.debug('JOIN-docproc-%d' % p_id)
        doc_processors[p_id].join_subprocess()
        logging.debug('TERMINATE-docproc-%d' % p_id)
    info_thread.join()
    logging.debug('JOIN-info')
    state_thread.join()
    logging.debug('JOIN-state')
except Exception as e:
    logging.critical(e)
    sys.stderr.write(str(e))
finally:
    logging.info('Runtime: %d s' % util.seconds_total(util.now() - time_start))
    urls_oversize_fp.close()
    save_state(domains, domain_distances, html_hashes, txt_hashes)
