#===============================================================================
#   SpiderLing -- A web spider/crawler for linguistic purpose.
#   Copyright (C) 2011-2015  Vit Suchomel
#   http://corpus.tools/wiki/SpiderLing
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================

import datetime, os, struct, subprocess
from time import sleep
from threading import Lock, Thread
from collections import deque
import config

TERM_ORDER = 'ORDER_66'
TERM_RESPONSE = 'TERMINATED'
MSG_NEW_URL = 'NEW_URL'
MSG_DEAD_DOMS = 'DEAD_DOMS'
MSG_ROBOT_LINES = 'ROBOT_LINES'
MSG_ROBOT_FAIL = 'ROBOT_FAIL'
MSG_ROBOT_EMPTY = ''
ROBOT_FLAG = 'ROBOT'
NOROBOT_FLAG = ''
PIPE_PATH = os.path.join(config.PIPE_DIR, 'pipe')

class SubProcess(object):
    PROCESS_SCRIPT = None
    PYTHON = 'pypy'
    IO_WAIT_DELAY = 0.2

    def __init__(self, name, logger=None):
        self._subproc = subprocess.Popen([self.PYTHON, self.PROCESS_SCRIPT],
            bufsize=0, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        self._name = name
        self._logger = logger
        self._stopped = False

    def get_send_data(self):
        return
    def process_recv_data(self, data):
        return

    def read(self):
        #read from subprocess until TERM_RESPONSE is received
        while True:
            data = recv_data(self._subproc.stdout)
            if data:
                if TERM_RESPONSE == data:
                    if self._logger:
                        self._logger.debug('TERM_RESPONSE from %s' % self._name)
                    break
                if not self._stopped:
                    self.process_recv_data(data)
            if self._stopped:
                sleep(self.IO_WAIT_DELAY)
        if self._logger:
            self._logger.info('Reading from %s terminated' % self._name)

    def write(self):
        #write to subprocess until TERM_ORDER is sent
        while True:
            if self._stopped:
                data = TERM_ORDER
            else:
                data = self.get_send_data()
            if data:
                try:
                    send_data(self._subproc.stdin, data)
                except IOError as e:
                    if self._logger:
                        self._logger.error(e)
                    break
                if self._stopped:
                    if self._logger:
                        self._logger.debug('Termination order sent to %s' % self._name)
                    self._subproc.stdin.flush()
                    break
            else:
                sleep(self.IO_WAIT_DELAY)
        if self._logger:
            self._logger.info('Writing to %s terminated' % self._name)

    def IO_threads_start(self):
        self._w_thread = Thread(target=self.write, name='%s-W' % self._name)
        self._w_thread.start()
        self._r_thread = Thread(target=self.read, name='%s-R' % self._name)
        self._r_thread.start()

    def IO_threads_join(self):
        self._w_thread.join()
        if self._logger:
            self._logger.debug('\t%s-W joined' % self._name)
        self._r_thread.join()
        if self._logger:
            self._logger.debug('\t%s-R joined' % self._name)

    def stop(self):
        self._stopped = True

    def join_subprocess(self):
        self._subproc.communicate()

class SafeSet(set):
    def __init__(self, items=()):
        super(SafeSet, self).__init__(items)
        self._lock = Lock()

    def add_(self, item):
        self._lock.acquire()
        try:
            self.add(item)
        finally:
            self._lock.release()

    def pop_(self):
        self.acquire()
        try:
            return self.pop()
        finally:
            self.release()

    def acquire(self):
        self._lock.acquire()

    def release(self):
        self._lock.release()

class ItemNotUniqueException(Exception):
    pass

class SafeUniqueSet(SafeSet):
    def add_unique(self, item):
        self._lock.acquire()
        try:
            if item in self:
                raise ItemNotUniqueException("'%s' is already in." % str(item))
            self.add(item)
        finally:
            self._lock.release()

    def __call__(self):
        return self.copy()

class SafeDict(dict):
    def __init__(self, *args):
        super(SafeDict, self).__init__(*args)
        self._lock = Lock()

    def acquire(self):
        self._lock.acquire()

    def release(self):
        self._lock.release()

    def set_(self, key, value):
        self.acquire()
        try:
            self[key] = value
        finally:
            self.release()

    def pop_(self, key, default=None):
        self.acquire()
        try:
            return self.pop(key, default)
        finally:
            self.release()

    def remove_(self, key):
        self.acquire()
        try:
            del(self[key])
        finally:
            self.release()

class SafeDeque(deque):
    def __init__(self, *args):
        super(SafeDeque, self).__init__(*args)
        self._lock = Lock()

    def acquire(self):
        self._lock.acquire()

    def release(self):
        self._lock.release()

class CheckFailException(Exception):
    pass

now = datetime.datetime.now

def seconds_total(timedelta_):
    return timedelta_.seconds + timedelta_.days * 24 * 3600

def seconds_left(time_start, duration):
    return max(0, duration - seconds_total(now() - time_start))

def timestamp():
    return datetime.datetime.now().strftime('%y%m%d%H%M%S')

def send_data(stream, data, flush=False, in_unicode=True):
    if in_unicode:
        data = data.encode('utf-8')
    stream.write('%d\n%s' % (len(data), data))
    if flush:
        stream.flush()

def recv_data(stream, in_unicode=True):
    result = ''
    #read data length
    data_length = stream.readline()
    if data_length:
        data_length = int(data_length)
        #read data
        read_bytes = 0
        read_data = []
        while read_bytes < data_length:
            next_data = stream.read(data_length - read_bytes)
            if next_data:
                read_data.append(next_data)
                read_bytes += len(next_data)
        result = ''.join(read_data)
    #convert to unicode
    if in_unicode:
        return unicode(result, 'utf-8')
    return result

def pack_ints(ints):
    if ints:
        return struct.pack('l' * len(ints), *ints)
    return ''

def unpack_ints(packed_ints):
    if packed_ints:
        return struct.unpack('l' * (len(packed_ints) / 8), packed_ints)
    return ()

def url_escape(s):
    return s.replace('"', '&quot;').replace('[', '%5B').replace(']', '%5D').replace('\\', '%5C')

def reset_dir(path):
    from shutil import rmtree
    from os import mkdir
    try:
        rmtree(path)
    except OSError as e:
        from errno import ENOENT
        if e.errno != ENOENT:
            raise
    mkdir(path)
