#===============================================================================
#   SpiderLing -- A web spider/crawler for linguistic purpose.
#   Copyright (C) 2011-2015  Vit Suchomel
#   http://corpus.tools/wiki/SpiderLing
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================

#===============================================================================
# SPIDERLING CONFIGURATION FILE
# Set the most important configuration options here
# Sections: general, crawling, domains, processing, debug
#===============================================================================


#== general ==

from math import log

#output and tmp file prefix
PREFIX = 'spiderling'
#directory for temporary named pipes
PIPE_DIR = '/tmp/%s' % PREFIX

#max crawling time in seconds (None for no limit)
MAX_RUN_TIME = None

#The number of downloaded data processing threads (process.py) to run,
#2 is enough for very small languages, 16 is enough for huge English crawls.
#Do not set to more than (<number of your processors> - 2).
#A single document processor can deal with approx. 1000 MB/hour raw data
#download speed (a rough estimate, depends on CPU).
DOC_PROCESSOR_COUNT = 2


#== crawling ==

#crawler identity
AGENT = '' #CHANGE e.g. to your project name
AGENT_URL = '+http://example.com/crawler.html' #CHANGE to your project url
USER_AGENT = 'Python-urllib/2.7 (%s)' % AGENT_URL
#A web browser identity can be used instead not to get rejected by some sites.
#Some sites use the '+' in AGENT_URL to reject bots => it is polite to keep it.
#The crawler's behaviour and possible masking is __your__ responsibility.

#https support, fall back to http if disabled
#ssl.SSLContext requires Python >= 2.7.9 (=> experimental, not tested thoroughly)
HTTPS_ENABLED = False

#http response size constraints
DOC_SIZE_MIN = 0.2*1024
DOC_SIZE_MAX = 5*1024*1024

#max number of open connections (max: ulimit -n), must be >> OPEN_AT_ONCE
MAX_OPEN_CONNS = 3000
#max number of connections opened at once
OPEN_AT_ONCE = 800
#abandon open connection when not responding (in seconds)
NO_ACTIVITY_TIMEOUT = 25

#max number of urls waiting to be downloaded
MAX_URL_QUEUE = 250000
#max number of documents waiting to be sent
MAX_DOC_QUEUE = 10000

#period of establishing a connection to the same IP .. 0.5 means 2/s
IP_CONN_INTERVAL = 0.5
#period of establishing a connection to the same domain .. 12.0 means 5/min
HOST_CONN_INTERVAL = 12.0
#The crawler's behaviour is __your__ responsibility, 
#setting these values too low will result in banning your IP by target sites
#and can lead to blocking your internet connection by your ISP!


#== domains ==

#min docs downloaded from a domain before yield rate threshold applies
MIN_DOCS_DOWNLOADED = 6
#min bytes downloaded from a domain before yield rate threshold applies
MIN_BYTES_DOWNLOADED = 256*1024
#max (soft limit) cleaned non-empty docs from a domain (None for no limit)
MAX_DOCS_CLEANED = None

"""
The yield rate threshold function sets the threshold to stop crawling a domain
when it is no more efficient.
"""
def yield_rate_threshold(doc_count):
    #standard threshold: 10: 0 %, 100: 1 %, 1000: 2 %, 10000: 3 %, 100000: 4 %
    return 0.01 * (log(doc_count, 10) - 1)
    #small languages threshold
    #return 0.005 * (log(doc_count, 10) - 1)

#The web domain data (paths within the domain) is the most memory consuming
#structure. A rough estimate is 500 MB - 5 GB per million domains, it depends
#on the number of urls within a stored domain. The more domains the crawler has
#to store, the more memory allocated. After reaching MAX_DOMS_READY domains,
#additional domains in state "ready" are written into domains_oversize file
#and never examined again. Therefore do not set MAX_DOMS_READY to a low value.
#Domains in state "new" (IP not determined yet) consume memory in the same way.
#The size of that queue is not limited. Therefore your free memory can be
#depleted by "new" domains despite you limited "ready" domains.

#max number of domains ready for crawling kept in memory
#1M is enough for small languages (expect ~2 GB RAM required),
#50M should be enough for huge English crawls (allocates up to ~200 GB RAM).
MAX_DOMS_READY = 2*1000*1000
#max domains selected for crawling at the same time
MAX_DOMS_SELECTED = 100*1000
#max time no data received from a domain in seconds
MAX_IDLE_TIME = 1800
#period of domain refresh (removing old, adding new, evaluating efficiency) in seconds
CHECK_DOMAINS_PERIOD = 600
#max domains with the same IP address waiting in the queue for the IP to become free
MAX_DOMS_WAITING_FOR_SINGLE_IP = 750
#max number of domains with the same IP selected for crawling at the same time
MAX_IP_REPEAT = 30

#Number of urls sent at once by the scheduler to the crawler. Decrease the
#minimum value to 1 if starting with less than 100 urls (even that may not help,
#flushing the I/O buffers would be needed). Increase the maximum value to allow
#wider domain coverage in the crawler's url queue. Must be < MAX_DOMS_SELECTED.
MIN_URL_SELECT = 50
MAX_URL_SELECT = 10000

#count of DNS resolving (host -> IP) threads to run
#1 is enough for most cases, 2 is enough for resolving many many domains
DNS_RESOLVERS = 1


#== processing ==

#Set the languages of documents to recognise, e.g. ['English', 'French'].
#Requires
# - a plaintext in that languages: ./util/lang_samples/English
# - a jusText model for that languages: ./justext/stoplists/English.txt
# - a chared model for that languages: ./chared/models/English
#The most similar language is selected (if below lang diff thresholds).
LANGUAGES = ['English']
#Set the languages of documents to accept, e.g. ['English'].
LANGUAGES_ACCEPT = ['English']
#max allowed difference to the lang model for document and paragraphs
#(0.0 = same as the model); it does not work well with paragraphs, the reported
#similarity is usually low
LANG_DIFF_THRESHOLD_DOC = 0.6
LANG_DIFF_THRESHOLD_PAR = 0.99
#disable encoding detection and force an encoding (e.g. 'utf-8'), None by default
FORCE_ENCODING = None
#Switch to unigram models (useful for Chinese, Japanese, Korean)
UNIGRAM_MODELS = False

#allowed top level domains regexp -- crawling is restricted to these, e.g.:
# - English + general web: '\.(?:uk|us|au|ca|com|org|net|info|biz|edu|name)$'
# - German: '\.(?:de|at|ch|com|org|net|info|eu|edu|name)$'
# - Czech: '\.cz'
# - no restriction: ''
from re import compile
ALLOWED_TLDS_RE = compile('')
#allowed domains to override restrictions in allowed_tlds_re
ALLOWED_EXTRA_DOMAINS = ()
#blacklist file of web domains, None by default
DOMAIN_BLACKLIST_FILE = None

#path to chared installation, file CHARED_PATH/models/LANGUAGE must exist
CHARED_PATH = 'chared'


#== debug ==

import logging
LOG_LEVEL = logging.INFO #DEBUG|INFO|WARNING|ERROR
#period of writing out some debug info in seconds, 0 to disable
INFO_PERIOD = 0
#period of saving state of the crawler (all domain data), 0 to disable
STATE_PERIOD = 0
