#===============================================================================
#   SpiderLing -- A web spider/crawler for linguistic purpose.
#   Copyright (C) 2011-2015  Vit Suchomel
#   http://corpus.tools/wiki/SpiderLing
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================

from threading import Thread
from time import sleep

import util
from util import config
from util.http import get_ip, NO_HOST, BAD_IP, ROBOT_PATH
from util.robotparser import RobotExclusionRulesParser

import logging

class Domain(object):
    __slots__ = ('RP_NONE', 'RP_INIT', 'RP_READY', 'RP_FAIL', 'RP_EMPTY',
        '_host', '_ip', '_bad', '_paths_hashes', '_paths_new',
        '_count_to_download', '_count_downloaded', '_count_cleaned',
        '_bytes_downloaded', '_bytes_cleaned', '_robot_lines', '_robot_parser',
        '_robot_state', '_last_action_time')

    RP_NONE = 0
    RP_INIT = 1
    RP_READY = 2
    RP_FAIL = 3
    RP_EMPTY = 4

    def __init__(self, host):
        self._host = host
        self._ip = None
        self._bad = False
        self._paths_hashes = util.SafeUniqueSet() #all paths' hashes
        self._paths_new = util.SafeSet() #paths waiting to be sent for download
        self._count_to_download = 0 #number of paths sent to be downloaded
        self._count_downloaded = 0 #number of paths successfuly downloaded (html)
        self._count_cleaned = 0 #number of paths successfuly cleaned (txt)
        self._bytes_downloaded = 0
        self._bytes_cleaned = 0
        self._robot_lines = ''
        self._robot_parser = RobotExclusionRulesParser()
        self._robot_state = self.RP_NONE
        self._last_action_time = util.now()

    def __cmp__(self, other):
        return cmp(self.get_cleaned_size_ratio(), other.get_cleaned_size_ratio())

    def get_host(self):
        return self._host

    def set_ip(self, ip):
        if ip in (NO_HOST, BAD_IP, None):
            self._bad = True
        self._ip = ip

    def get_ip(self):
        return self._ip
    
    def is_bad(self):
        return self._bad

    def set_robot_state(self, state):
        if not state in (self.RP_NONE, self.RP_INIT, self.RP_READY, 
                         self.RP_FAIL, self.RP_EMPTY):
            raise KeyError('Invalid robot parser state')
        self._robot_state = state
        if self.RP_FAIL == self._robot_state:
            self._bad = True
    
    def is_robot_ready(self):
        return self.RP_READY == self._robot_state

    def set_robot_lines(self, lines):
        if not lines:
            self.set_robot_state(self.RP_EMPTY)
        elif util.MSG_ROBOT_FAIL == lines:
            self.set_robot_state(self.RP_FAIL)
        else:
            self._robot_lines = lines
            self._robot_parser.parse(lines)
            self.set_robot_state(self.RP_READY)
            #filter out paths not allowed for robots
            paths_new_tmp = self._paths_new
            paths_new_tmp.acquire()
            try:
                self._paths_new = util.SafeSet()
                for path, schema in paths_new_tmp:
                    if self.is_robot_allowed(path):
                        self._paths_new.add((path, schema))
            finally:
                paths_new_tmp.release()

    def is_robot_allowed(self, path):
        if self.RP_READY == self._robot_state:
            try:
                return self._robot_parser.is_allowed(config.AGENT, path)
            except Exception:
                logging.warning('Robot parser failed for domain %s, disallowing access to %s.'
                    % (self._host.encode('utf-8'), path))
                return False
        if self.RP_EMPTY == self._robot_state:
            return True
        return False

    def add_new_path(self, path, schema):
        #check path has not been added yet
        try:
            self._paths_hashes.add_unique(hash(path))
        except util.ItemNotUniqueException:
            return
        #check path is allowed for robots
        if self._robot_state in (self.RP_NONE, self.RP_INIT) \
                or self.is_robot_allowed(path):
            self._paths_new.add_((path, schema))
        self._last_action_time = util.now()

    def get_path_to_download(self):
        if self._bad:
            return
        if self.RP_NONE == self._robot_state:
            self.set_robot_state(self.RP_INIT)
            return (ROBOT_PATH, 'http', util.ROBOT_FLAG)
        if self._robot_state in (self.RP_EMPTY, self.RP_READY):
            try:
                path, schema = self._paths_new.pop_()
            except KeyError:
                return
            self._count_to_download += 1
            return (path, schema, util.NOROBOT_FLAG)

    def add_size_downloaded(self, size):
        self._count_downloaded += 1
        self._bytes_downloaded += size
        self._last_action_time = util.now()

    def add_size_cleaned(self, size):
        self._count_cleaned += 1
        self._bytes_cleaned += size

    def get_count_cleaned(self):
        return self._count_cleaned

    def get_cleaned_size_ratio(self):
        if 1 > self._bytes_downloaded:
            return 0.0
        return float(self._bytes_cleaned) / self._bytes_downloaded

    def is_efficient(self):
        #not enough data/files got
        if config.MIN_BYTES_DOWNLOADED > self._bytes_downloaded or \
                config.MIN_DOCS_DOWNLOADED > self._count_downloaded:
            return True
        #evaluate & compare with the threshold:
        size_ratio_threshold = config.yield_rate_threshold(self._count_downloaded)
        cleaned_size_ratio = self.get_cleaned_size_ratio()
        result = size_ratio_threshold < cleaned_size_ratio
        if not result:
            self._bad = True
        return result

    def is_idle(self):
        if not self._paths_new and 1 > util.seconds_left(
                self._last_action_time, config.MAX_IDLE_TIME):
            return True

    def new_paths_count(self):
        return len(self._paths_new)

    def get_last_action_time(self):
        return self._last_action_time

    def update_last_action_time(self):
        self._last_action_time = util.now()

    def save_to_string(self):
        p1 = (u'%s\n%s\n%s\n%d\n%d\n%d\n%d\n%d\n%d\n' % (
            self._host, self._ip, self._bad, self._robot_state, 
            self._bytes_downloaded, self._bytes_cleaned, 
            self._count_to_download, self._count_downloaded, 
            self._count_cleaned)).encode('utf-8')
        p2 = self._robot_lines.encode('utf-8')
        p3 = u'\n'.join(map(u'\t'.join, self._paths_new)).encode('utf-8')
        p4 = util.pack_ints(self._paths_hashes)
        return '\n~\n'.join((p1, p2, p3, p4))

    def info(self):
        return (u'%s\t%s\t%s\t%d\t%d\t%d\t%d\t%d\t%d' % (
            self._host, self._ip, self._bad, self._robot_state, 
            self._bytes_downloaded, self._bytes_cleaned, 
            self._count_to_download, self._count_downloaded, 
            self._count_cleaned)).encode('utf-8')

    def load_from_string(self, string):
        p1, p2, p3, p4 = string.split('\n~\n', 3)
        host, ip, bad, robot_state, bytes_downloaded, bytes_cleaned, \
            count_to_download, count_downloaded, count_cleaned = \
            p1.split('\n', 8)
        self._host = unicode(host, 'utf-8')
        if 'None' != ip:
            self.set_ip(ip)
        if 'True' == bad:
            self._bad = True
        else:
            self._bad = False
            self._bytes_downloaded = int(bytes_downloaded)
            self._bytes_cleaned = int(bytes_cleaned)
            self._count_to_download = int(count_to_download)
            self._count_downloaded = int(count_downloaded)
            self._count_cleaned = int(count_cleaned)
            self.set_robot_lines(unicode(p2, 'utf-8'))
            self.set_robot_state(int(robot_state))
            for p3_line in unicode(p3, 'utf-8').splitlines():
                self._paths_new.add(p3_line.split('\t'))
            self._paths_hashes.update(set(util.unpack_ints(p4)))

    def get_stats(self):
        return (self._host, self._ip, self._bad, self._robot_state, \
            len(self._paths_new), len(self._paths_hashes), \
            self._count_to_download, self._count_downloaded, \
            self._count_cleaned, self._bytes_downloaded, self._bytes_cleaned, \
            self.get_cleaned_size_ratio())

class DomainQueue():
    BAD_DOMS_FILE = '%s.domains_bad' % config.PREFIX #where to put bad (emptied etc.) domains
    OVERSIZE_DOMS_FILE = '%s.domains_oversize' % config.PREFIX #where to put possibly good domains but not meeting limits above

    def __init__(self, urls_oversize_fp):
        self._host_dom_map = {}
        self._doms_ip_resolve = [] #domains waiting for IP address
        self._doms_new = util.SafeDict()
        self._doms_ready = {}
        self._doms_selected = util.SafeDict()
        self._doms_bad = {}
        self._ips_selected = {}
        self._good_urls_buffer = set()
        self._terminate = False
        self._urls_oversize_fp = urls_oversize_fp
        self._doms_dead = util.SafeSet()
        #start threads resolving new domains' IP addresses
        for i in range(config.DNS_RESOLVERS):
            Thread(target=self._dns_resolve, name='dom-DNS%d' % i).start()

    def __getitem__(self, host):
        return self._host_dom_map[host]
    
    def start(self):
        #start a thread periodically checking domains
        Thread(target=self._check_domains_loop, name='dom-check').start()

    def _dns_resolve(self):
        NO_DOM_DELAY = 10
        while not self._terminate:
            try:
                domain = self._doms_ip_resolve.pop()
            except IndexError:
                sleep(NO_DOM_DELAY)
            else:
                domain.set_ip(get_ip(domain.get_host())) #blocks

    def _check_domains_loop(self):
        sleep(20)
        self.check_domains()
        sleep(30)
        self.check_domains()
        period_start = util.now()
        while not self._terminate:
            if 1 > util.seconds_left(period_start, config.CHECK_DOMAINS_PERIOD):
                self.check_domains()
                period_start = util.now()
            sleep(60)

    def check_domains(self):
        logging.info('check_domains start')

        #new -> ready | bad
        ready_new = bad_new = 0
        self._doms_new.acquire()
        try:
            hosts = self._doms_new.keys()
            for host in hosts:
                domain = self._doms_new[host]
                if domain.is_bad():
                    self._doms_bad[host] = domain
                    del(self._doms_new[host])
                    bad_new += 1
                elif domain.get_ip():
                    self._doms_ready[host] = domain
                    del(self._doms_new[host])
                    ready_new += 1
        finally:
            self._doms_new.release()

        #selected bad/inefficient/max_docs -> bad; selected idle -> ready
        ready_idle = bad_bad = bad_eff = bad_max_docs = 0
        self._doms_selected.acquire()
        try:
            hosts = self._doms_selected.keys()
            for host in hosts:
                domain = self._doms_selected[host]
                remove = False
                if domain.is_bad():
                    remove = True
                    bad_bad += 1
                elif domain.is_idle():
                    self._doms_ready[host] = domain
                    del(self._doms_selected[host])
                    self._ips_selected[domain.get_ip()] -= 1
                    ready_idle += 1
                elif not domain.is_efficient():
                    remove = True
                    bad_eff += 1
                elif config.MAX_DOCS_CLEANED and \
                        domain.get_count_cleaned() >= config.MAX_DOCS_CLEANED:
                    remove = True
                    bad_max_docs += 1
                if remove:
                    self._doms_bad[host] = domain
                    del(self._doms_selected[host])
                    self._ips_selected[domain.get_ip()] -= 1
        finally:
            self._doms_selected.release()

        #bad -> remove (bad)
        with open(self.BAD_DOMS_FILE, 'a') as fp:
            for host, domain in self._doms_bad.iteritems():
                fp.write('%s\n~DOM\n' % domain.save_to_string())
                self._host_dom_map[host] = None
                self._doms_dead.add_(host)
        self._doms_bad = {}

        #ready -> selected
        selected = wait_empty = 0
        self._doms_selected.acquire()
        wait_ip = 0
        try:
            max_select = config.MAX_DOMS_SELECTED - len(self._doms_selected)
            if 0 < max_select:
                #sort domains by new path count (desc)
                doms_ready_sorted = {}
                for dom in sorted(self._doms_ready.values(), lambda x,y: 
                        cmp(y.new_paths_count(), x.new_paths_count())):
                    doms_ready_sorted[dom.get_host()] = dom
                #make domains with most new paths selected
                #the selected IPs are distributed evenly up to config.MAX_IP_REPEAT
                domains_to_omit = []
                ip_repeat = 1
                while config.MAX_IP_REPEAT >= ip_repeat and max_select > selected:
                    if domains_to_omit:
                        for host in domains_to_omit:
                            del(doms_ready_sorted[host])
                        domains_to_omit = []
                    wait_ip = 0
                    for host, domain in doms_ready_sorted.iteritems():
                        #do not select doms without new paths
                        if 1 > domain.new_paths_count():
                            wait_empty += 1
                            domains_to_omit.append(host)
                            continue
                        #do not select doms with IP selected too many times
                        ip = domain.get_ip()
                        ip_select_count = self._ips_selected.get(ip, 0)
                        if ip_select_count >= ip_repeat:
                            wait_ip += 1
                        else:
                            self._doms_selected[host] = domain
                            del(self._doms_ready[host])
                            self._ips_selected[ip] = ip_select_count + 1
                            domain.update_last_action_time()
                            domains_to_omit.append(host)
                            selected += 1
                            if selected >= max_select:
                                break
                    ip_repeat += 1
        finally:
            self._doms_selected.release()

        #ready -> remove (oversize)
        oversize = 0
        max_remove = max(0, len(self._doms_ready) - config.MAX_DOMS_READY)
        if max_remove:
            #get ip count
            ips = {}
            for domain in self._doms_ready.itervalues():
                ip = domain.get_ip()
                ips[ip] = ips.get(ip, 0) + 1
            #sort domains by new path count (asc) and update time (asc)
            doms_ready_sorted = sorted(self._doms_ready.values(), 
                lambda x,y: cmp(x.new_paths_count(), y.new_paths_count()) 
                or cmp(x.get_last_action_time(), y.get_last_action_time()))
            doms_ready_sorted_rest = []
            #remove domains
            with open(self.OVERSIZE_DOMS_FILE, 'a') as fp:
                #remove domains with many other domains waiting for the same IP
                for domain in doms_ready_sorted:
                    ip = domain.get_ip()
                    if config.MAX_DOMS_WAITING_FOR_SINGLE_IP < ips[ip]:
                        host = domain.get_host()
                        fp.write('%s\n~DOM\n' % domain.save_to_string())
                        self._host_dom_map[host] = None
                        del(self._doms_ready[host])
                        ips[ip] -= 1
                        oversize += 1
                        if oversize >= max_remove:
                            break
                    elif len(doms_ready_sorted_rest) < max_remove - oversize:
                        doms_ready_sorted_rest.append(domain)
                #remove low path count and not recently modified domains
                remains_to_remove = max(0, max_remove - oversize)
                for domain in doms_ready_sorted_rest[:remains_to_remove]:
                    host = domain.get_host()
                    fp.write('%s\n~DOM\n' % domain.save_to_string())
                    self._host_dom_map[host] = None
                    del(self._doms_ready[host])
                    oversize += 1

        logging.info('check_domains end, %d ready (%d new, %d idle), %d selected, %d waiting '
                '(%d IP, %d empty), %d oversize, %d bad (%d new, %d robots, %d eff, %d max docs)' %
            (ready_new + ready_idle, ready_new, ready_idle, selected,
            wait_ip + wait_empty, wait_ip, wait_empty, oversize,
            bad_new + bad_bad + bad_eff + bad_max_docs, bad_new, bad_bad, bad_eff, bad_max_docs))

    def add(self, host, path, schema):
        #update the domain
        try:
            domain = self._host_dom_map[host]
        except KeyError:
            #create a new domain
            domain = Domain(host)
            self._host_dom_map[host] = domain
            domain.add_new_path(u'/', schema)
            self._doms_new.acquire()
            try:
                self._doms_new[host] = domain
            finally:
                self._doms_new.release()
            self._doms_ip_resolve.append(domain)
        if domain:
            domain.add_new_path(path.strip(), schema)
        else:
            #adding a url to a domain which has already been dumped => dump the url
            self._urls_oversize_fp.write('%s\t%s\n' % (host.encode('utf-8'), path.encode('utf-8')))

    def get_good_urls(self):
        result = []
        #get urls from good_urls_buffer first
        got_count = 0
        while self._good_urls_buffer and config.MAX_URL_SELECT > got_count:
            result.append(self._good_urls_buffer.pop())
            got_count += 1
        #get next round of urls from domains
        if config.MAX_URL_SELECT > got_count:
            self._doms_selected.acquire()
            try:
                for host, domain in self._doms_selected.iteritems():
                    try:
                        path, schema, robot_flag = domain.get_path_to_download()
                    except (TypeError, ValueError):
                        continue
                    url_tuple = (schema, host, path, domain.get_ip(), robot_flag)
                    got_count += 1
                    #add url to result, add overflow url to good_urls_buffer
                    if config.MAX_URL_SELECT > got_count:
                        result.append(url_tuple)
                    else:
                        self._good_urls_buffer.add(url_tuple)
            finally:
                self._doms_selected.release()
            #update good_urls_buffer and return [] if min count not reached
            if config.MIN_URL_SELECT > got_count and 0 < got_count:
                self._good_urls_buffer.update(result)
                return []
        return result

    def get_doms_dead(self):
        self._doms_dead.acquire()
        result = self._doms_dead.copy()
        self._doms_dead.clear()
        self._doms_dead.release()
        return result

    def terminate(self):
        self._terminate = True

    def get_stats(self):
        return (len(self._doms_new), len(self._doms_ready), 
            len(self._doms_selected), len(self._doms_bad))

    def save_to_file(self, fp):
        for domain in self._host_dom_map.values():
            if domain:
                #!TODO rewrite, use chunk size instead of end of record marker (~DOM)
                fp.write('%s\n~DOM\n' % domain.save_to_string())
    
    def load_from_file(self, fp):
        domains_loaded = 0
        dom_string_list = []
        for line in fp:
            if line.startswith('~DOM'):
                #load a domain
                dom_string = ''.join(dom_string_list)[:-1]
                dom_string_list = []
                domain = Domain(None)
                try:
                    domain.load_from_string(dom_string)
                except Exception:
                    logging.error('Could not load domain from string: "%s"' % dom_string)
                    continue
                domains_loaded += 1
                if 0 == domains_loaded % 10000:
                    logging.debug('%d doms loaded' % domains_loaded)
                #overwrite a domain with the same hostname if exists
                host = domain.get_host()
                if host in self._host_dom_map:
                    try:
                        del(self._doms_bad[host])
                    except KeyError:
                        pass
                    try:
                        del(self._doms_ready[host])
                    except KeyError:
                        pass
                    try:
                        del(self._doms_new[host])
                    except KeyError:
                        pass
                    try:
                        self._doms_ip_resolve.remove(domain)
                    except ValueError:
                        pass
                #put the domain into the corresponding structure
                if domain.is_bad():
                    self._host_dom_map[host] = None
                    self._doms_bad[host] = domain
                else:
                    self._host_dom_map[host] = domain
                    if domain.get_ip():
                        self._doms_ready[host] = domain
                    else:
                        self._doms_new[host] = domain
                        self._doms_ip_resolve.append(domain)
            else:
                dom_string_list.append(line)
        logging.info('%d domains loaded' % domains_loaded)
