#===============================================================================
#   SpiderLing -- A web spider/crawler for linguistic purpose.
#   Copyright (C) 2011-2015  Vit Suchomel
#   http://corpus.tools/wiki/SpiderLing
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================

#!TODO Content-Encoding: gzip

from errno import EINPROGRESS
from os.path import dirname
from posixpath import normpath
import re
import socket, ssl
import urlparse
import util
from util import config

import logging

HTTP_REDIRECT_CODES = [301, 302, 303, 307]
ROBOT_PATH = '/robots.txt'
CHUNK_SIZE = 100*1024
NO_HOST = 'NO_HOST'
BAD_IP = 'BAD_IP'

if config.HTTPS_ENABLED:
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
    ssl_context.verify_mode = ssl.CERT_NONE
    ssl_context.check_hostname = False
    ssl_context.load_default_certs()

def get_ip(host):
    host = host.encode('utf-8')
    #get address info -- IPv4 or IPv6, blocking
    try:
        addrinfo = socket.getaddrinfo(host, 80, socket.AF_INET)
    except (socket.gaierror, socket.error):
        return NO_HOST
    #get IP address
    try:
        ip = addrinfo[0][4][0]
    except IndexError:
        return BAD_IP
    return ip

def url_split(url):
    parts = urlparse.urlsplit(url)
    path, query = parts.path, parts.query
    full_path = ('%s?%s' % (path, query)) if query else path
    return (parts.scheme, parts.netloc, full_path)

def url_join(schema, host, full_path):
    if not full_path.startswith('/'):
        full_path = '/%s' % full_path
    return '%s://%s%s' % (schema, host, full_path)

url_scheme_re = re.compile('([^:/]+)://')
url_with_no_path_re = re.compile('://[^/]+$')
def url_join_rel_norm(base_url, new_url):
    #resolve relative paths, join with base url, normalise (no trailing slash, no fragment)
    base_url, new_url = base_url.replace(' ', ''), new_url.replace(' ', '')
    if url_with_no_path_re.search(base_url):
        base_url = base_url + '/'
    if new_url.startswith('//'):
        m = url_scheme_re.match(base_url)
        new_url = u'%s:%s' % (m.group(1) if m else 'http', new_url)
    url_parts = urlparse.urlparse(new_url)
    if url_parts.scheme == '' and url_parts.netloc == '': #relative path
        start_slash = '/' if not new_url.startswith('/') else ''
        new_url = '%s%s%s' % (dirname(base_url), start_slash, new_url)
        if new_url.endswith('/'):
            new_url = new_url[:-1]
        url_parts = urlparse.urlparse(new_url)
    new_path = normpath(url_parts.path)
    if new_path in ('.', '/'):
        new_path = ''
    return urlparse.urlunsplit((url_parts.scheme, url_parts.netloc, new_path, url_parts.query, ''))

class HTTPException(Exception):
    pass
class IncorrectDataChunkError(HTTPException):
    pass
class ConnectionException(Exception):
    pass

class IncorrectResponseError(HTTPException):
    def __init__(self, message, remote_fault=False):
        super(IncorrectResponseError, self).__init__(message)
        #a flag saying we are 99.9 % sure the error is caused by the remote 
        # communication partner, not by an inability of this library
        # /we are only 90 % sure in other cases :-)/
        self._remote_fault = remote_fault 

    def is_remote_fault(self):
        return self._remote_fault

class RedirectException(HTTPException):
    def __init__(self, redirect_url):
        super(Exception, self).__init__()
        self._redirect_url = redirect_url
    
    def get_redirect_url(self):
        return self._redirect_url

    def __str__(self):
        return 'Redirect to %s' % self.get_redirect_url()


def build_request(host, path):
    http11_request_fields = ['GET %s HTTP/1.1' % path.encode('utf-8'),
                             'Host: %s' % host.encode('utf-8'),
                             'User-Agent: %s' % config.USER_AGENT,
                             #'Accept: text/html',
                             'Connection: close',
                             '', '']
    return '\r\n'.join(http11_request_fields)

class StringList():
    def __init__(self):
        self._list = []
        self._length = 0

    def __str__(self):
        return ''.join(self._list)

    def __len__(self):
        return self._length

    def append(self, string):
        self._list.append(string)
        self._length += len(string)

class Connection():
    def __init__(self, url, schema, host, path, ip, is_robot):
        self._url = url
        self._host = host
        self._path = path
        self._ip = ip
        self._is_robot = is_robot #a connection meant to get robots.txt
        self._https = config.HTTPS_ENABLED and (schema == 'https')
        #create a socket, wrap in SSL/TLS in case of https
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if self._https:
            self._socket = ssl_context.wrap_socket(self._socket, server_hostname=host)
        self._data_read = StringList()
        self.update_last_action()
        self._connect_time = None

    def get_url(self):
        return self._url
    
    def get_host(self):
        return self._host

    def get_path(self):
        return self._path

    def get_ip(self):
        return self._ip

    def fileno(self):
        return self._socket.fileno()

    def is_robot(self):
        return self._is_robot

    def get_data_read(self):
        return self._data_read

    def connect(self):
        self.update_last_action()
        #non-blocking connect
        self._socket.setblocking(0)
        logging.info('CONN\t%s\t%s\t%s' % (self._ip.encode('utf-8'), self._host.encode('utf-8'),
            's' if self._https else ''))
        try:
            self._socket.connect((self._ip, 443 if self._https else 80))
        except socket.error as e:
            if EINPROGRESS != e.errno: #115 => ok (operation in progress)
                raise ConnectionException('Error while connecting: %s' % unicode(e).encode('utf-8'))
        self._connect_time = util.timestamp()

    def get_connect_time(self):
        return self._connect_time

    def handle_write(self):
        #check connection errors (the socket has to be connected)
        if 0 != self._socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR):
            self.close_connection()
            raise ConnectionException('Socket error while connecting.')
        #send message
        try:
            self._socket.sendall(build_request(self._host, self._path))
        except (socket.error, socket.gaierror):
            self.close_connection()
            raise ConnectionException('Socket error while sending data')
        self.update_last_action()

    def handle_read(self): 
        #raises ConnectionException, util.CheckFailException, IncorrectResponseError,
        # IncorrectDataChunkError, RedirectException
        #try reading data from the socket
        try:
            data = self._socket.recv(CHUNK_SIZE)
        except socket.error:
            #Connection reset by peer, Connection timed out,...
            self.close_connection()
            raise ConnectionException('Socket error while reading data')
        #some data read this time
        if data:
            self.update_last_action()
            #examine the HTTP header for Content-Length
            if not self._data_read:
                #first data chunk read, assuming the HTTP header is here
                s = re.search('[Cc]ontent-[Ll]ength:\s*(\d*)', data)
                if s and s.group(1):
                    data_len = int(s.group(1))
                    if config.DOC_SIZE_MAX < data_len:
                        raise ConnectionException('Content length not in range')
            #add read data
            self._data_read.append(data)
            #too much data to read => close the connection
            if config.DOC_SIZE_MAX < len(self._data_read) and not self._is_robot:
                self.close_connection()
                raise ConnectionException('Too much data to read')
            #exit but keep the connection to allow future reading
            return (True, ())
        try:
            #no data read this time, some data read earlier
            if self._data_read:
                #process all data obtained through this socket
                content_type, header, body = process_response(
                    str(self._data_read), self._url, self._is_robot) #raises exceptions
                if not self._is_robot:
                    post_checks(body, content_type)
                #exit, do not keep the connection
                return (False, (header, body))
            #no data read ever
            return (False, ())
        finally:
            self.close_connection()

    def close_connection(self):
        try:
            self._socket.close()
        except socket.error:
            pass

    def update_last_action(self):
        self.last_action_t = util.now()

    def check_last_action(self):
        fail = config.NO_ACTIVITY_TIMEOUT < util.seconds_total(util.now() - self.last_action_t)
        if fail:
            self.close_connection()
        return not fail

def chunked_data(data):
    #join data chunks
    chunk_start = 0
    data_length = len(data)
    result_data_list = []
    while chunk_start < data_length:
        m = re.match('\s*([0-9a-f]+)', data[chunk_start:], re.I or re.S)
        try:
            chunk_length = int(m.group(1), 16)
        except AttributeError, IndexError:
            raise IncorrectDataChunkError()
        if 0 == chunk_length:
            break
        chunk_start = data.find('\r\n', chunk_start) + 2
        chunk_end = chunk_start + chunk_length
        if data_length < chunk_end:
            raise IncorrectDataChunkError()
        chunk = data[chunk_start:chunk_end]
        if len(chunk) != chunk_length:
            raise IncorrectDataChunkError()
        result_data_list.append(chunk)
        chunk_start = chunk_end + 1
    return ''.join(result_data_list)

def header_1st_line(first_line):
    #process 1st line of a HTTP/1.1 response
    if not first_line:
        raise IncorrectResponseError('Empty first line')
    try:
        protocol, code, msg = re.split('\s+', first_line, maxsplit=2)
        code = int(code)
    except ValueError:
        raise IncorrectResponseError('Incorrect first line format')
    return (protocol, code, msg)

def process_response(http_response, url, check_valid_robot):
    #split a HTTP/1.1 response into header and body
    try:
        header, body = re.split('(?s)\r?\n\r?\n', http_response, maxsplit=1)
    except ValueError:
        header, body = (http_response, '')
    try:
        header_lines = re.split('\r\n', header)
        header_dict = {'_response': header_lines.pop(0)}
        header_dict['_protocol'], header_dict['_code'], header_dict['_msg'] = \
            header_1st_line(header_dict['_response'])
        for header_line in header_lines:
            try:
                key, val = re.split('\s*\:\s*', header_line, maxsplit=1)
            except ValueError:
                continue
            try:
                header_dict[key.lower()] = val
            except UnicodeDecodeError as e:
                logging.debug('UnicodeDecodeError:\n%s' % e)
                pass
    except Exception as e:
        raise IncorrectResponseError(unicode(e))

    #HTTP features
    #redirection
    if header_dict['_code'] in HTTP_REDIRECT_CODES:
        try:
            redirect_url = unicode(header_dict['location'], 'utf-8')
        except UnicodeDecodeError:
            raise IncorrectResponseError(
                'Cannot transform a redirect url to unicode from utf-8')
        except KeyError:
            raise IncorrectResponseError(
                'Missing redirect location', remote_fault=True)
        try:
            redirect_url = url_join_rel_norm(url, redirect_url)
        except ValueError:
            raise IncorrectResponseError(
                'Incorrect URL (urlparse fails)', remote_fault=True)
        raise RedirectException(redirect_url)
    #response not "200 OK"
    if 200 != header_dict['_code']:
        raise IncorrectResponseError(
            'Response code is not 200', remote_fault=True)
    #chunked transfer encoding
    if 'chunked' == header_dict.get('transfer-encoding'):
        body = chunked_data(body)
    #check the file (supposed to be robots.txt) is not a html file
    if check_valid_robot:
        body_lower = body.lower() 
        if '<html' in body_lower or '<body' in body_lower:
            raise IncorrectResponseError(
                'robots.txt replaced by an html document', remote_fault=True)

    #return content type, header, body
    content_type = header_dict.get('content-type')
    return (content_type, header, body)

def post_checks(html_text, content_type):
    #check content type
    if not content_type or ('text/html' not in content_type 
                            and 'text/plain' not in content_type): #? 
        raise util.CheckFailException("Content type %s rejected" % content_type)
    #check document size
    if not html_text or config.DOC_SIZE_MIN > len(html_text):
        raise util.CheckFailException("Document too small: %d" % len(html_text))
